#include "ResultFile.h"
#include "VotingSystem.h"
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>

OpenResultFile::OpenResultFile() {
}
int OpenResultFile::create( char* path, VotingSystem** s, int ns,
    u_int32_t* voterStepsIn,
    u_int32_t voterStepsEntriesIn,
    u_int32_t* choiceStepsIn,
    u_int32_t choiceStepsEntriesIn,
    float* errorStepsIn,
    u_int32_t errorStepsEntriesIn
 ) {
    int err;
    u_int32_t pos = 0;
    size_t sizeOfData;

    voterStepsEntries = voterStepsEntriesIn;
    choiceStepsEntries = choiceStepsEntriesIn;
    errorStepsEntries = errorStepsEntriesIn;
    systemsEntries = ns;

    votingSystemArrayToNameBlock( &names, s, ns );

    fd = ::open( path, O_RDWR | O_CREAT, 0777 );
    if ( fd < 0 ) {
	perror("open");
	return fd;
    }
    estep = 8 + systemsEntries * sizeof(struct SystemResult);
    cstep = estep * errorStepsEntries;
    vstep = cstep * choiceStepsEntries;
    sizeOfData = sizeof(Result) * vstep * voterStepsEntries;
    fprintf( stderr, "estep %d\tcstep %d\tvstep %d\nsizeOfData %d\n", estep, cstep, vstep,
	(unsigned int)sizeOfData );
    rfSize = sizeOfData +
	sizeof(u_int32_t) * (choiceStepsEntries + voterStepsEntries) +
	sizeof(float) * errorStepsEntries +
	sizeof(ResultFile) +
	names.blockLen + 64 /* alignment fudge factor */;
#ifdef F_PREALLOCATE
    {
    fstore_t size;
    size.fst_flags = F_ALLOCATEALL;
    size.fst_posmode = F_PEOFPOSMODE;
    size.fst_offset = 0;
    size.fst_length = rfSize;
    err = fcntl( fd, F_PREALLOCATE, &size );
    if ( err == -1 ) {
	fprintf( stderr, "tried to allocate %d bytes and failed\n", (unsigned int)rfSize );
	perror("fcntl(F_PREALLOCATE)");
	return err;
    }
    }
#endif
    {
	char* zero;
	size_t sizeleft = rfSize;
	size_t ps = getpagesize();
	zero = (char*)malloc( ps );
	bzero( zero, ps );
	while ( sizeleft > ps ) {
	    write( fd, zero, ps );
	    sizeleft -= ps;
	}
	if ( sizeleft )
	    write( fd, zero, sizeleft );
	free( zero );
    }
#if 0
    ::close( fd );
    fd = ::open( path, O_RDWR, 0777 );
#endif
    rf = (ResultFile*)mmap( NULL, rfSize, PROT_READ|PROT_WRITE, MAP_FILE | MAP_SHARED, fd, 0 );
    if ( rf == (ResultFile*)-1 ) {
	perror("mmap");
	return -1;
    }
    pos = sizeof(ResultFile);
    rf->dataStartOffset = pos;
    pos += sizeOfData;
    rf->voterStepsOffset = pos;
    rf->voterStepsEntries = voterStepsEntries;
    voterSteps = (u_int32_t*)(((caddr_t)rf) + rf->voterStepsOffset);
    memcpy( voterSteps, voterStepsIn, sizeof(u_int32_t) * voterStepsEntries );
    pos += voterStepsEntries * sizeof(u_int32_t);
    rf->choiceStepsOffset = pos;
    rf->choiceStepsEntries = choiceStepsEntries;
    choiceSteps = (u_int32_t*)(((caddr_t)rf) + rf->choiceStepsOffset);
    memcpy( choiceSteps, choiceStepsIn, sizeof(u_int32_t) * choiceStepsEntries );
    pos += choiceStepsEntries * sizeof(u_int32_t);
    rf->errorStepsOffset = pos;
    rf->errorStepsEntries = errorStepsEntries;
    errorSteps = (float*)(((caddr_t)rf) + rf->errorStepsOffset);
    memcpy( errorSteps, errorStepsIn, sizeof(u_int32_t) * errorStepsEntries );
    pos += errorStepsEntries * sizeof(u_int32_t);
    rf->systemsOffset = pos;
    rf->systemsEntries = systemsEntries;

    {
	void* block;
	block = ((caddr_t)rf) + rf->systemsOffset;
	memcpy( block, names.block, names.blockLen );
    }
    {
	u_int32_t i, j, nR;
	Result* r;
	caddr_t rp = (((caddr_t)rf) + rf->dataStartOffset);
	nR = choiceStepsEntries * errorStepsEntries * voterStepsEntries;
	for ( i = 0; i < nR; i++ ) {
	    r = (Result*)rp;
	    r->trials = 0;
	    for ( j = 0; j < systemsEntries; j++ ) {
		r->systems[j].meanHappiness = 0;
		r->systems[j].reliabilityStd = 0;
		r->systems[j].consensusStdAvg = 0;
	    }
	    rp += estep;
	}
    }

#if 01
    err = flush();
#else
    err = msync( rf, rfSize, MS_SYNC );
    if ( err ) {
	perror("msync");
    }
#endif
    return err;
}
int OpenResultFile::open( char* path ) {
    struct stat sb;
    int err;

    fd = ::open( path, O_RDWR );
    if ( fd < 0 ) {
	perror("open");
	return fd;
    }
    err = fstat( fd, &sb );
    if ( err != 0 ) {
	perror("fstat");
	return err;
    }
    rfSize = sb.st_size;
    rf = (ResultFile*)mmap( NULL, rfSize, PROT_READ|PROT_WRITE, MAP_FILE | MAP_SHARED, fd, 0 );
    if ( rf == (ResultFile*)-1 ) {
	perror("mmap");
	return -1;
    }
    voterStepsEntries = rf->voterStepsEntries;
    choiceStepsEntries = rf->choiceStepsEntries;
    errorStepsEntries = rf->errorStepsEntries;
    systemsEntries = rf->systemsEntries;
    estep = 8 + systemsEntries * sizeof(struct SystemResult);
    cstep = estep * errorStepsEntries;
    vstep = cstep * choiceStepsEntries;
    data = (((caddr_t)rf) + rf->dataStartOffset);
    voterSteps = (u_int32_t*)(((caddr_t)rf) + rf->voterStepsOffset);
    choiceSteps = (u_int32_t*)(((caddr_t)rf) + rf->choiceStepsOffset);
    errorSteps = (float*)(((caddr_t)rf) + rf->errorStepsOffset);
    names.block = (char*)(((caddr_t)rf) + rf->systemsOffset);
    names.blockLen = rfSize - rf->systemsOffset;
    parseNameBlock( &names );

    return 0;
}

int OpenResultFile::flush() {
    int err;
    err = msync( rf, rfSize, MS_SYNC );
    if ( err ) perror("msync");
    return err;
}
int OpenResultFile::close() {
    flush();
    munmap( rf, rfSize );
    ::close( fd );
}

int OpenResultFile::votersToVStep( u_int32_t numv ) {
    int vstep = -1;
    unsigned int i;
    for ( i = 0; i < voterStepsEntries; i++ ) {
	if ( numv == voterSteps[i] ) {
	    vstep = i;
	    break;
	}
    }
    return vstep;
}
int OpenResultFile::choicesToCStep( u_int32_t numc ) {
    int cstep = -1;
    unsigned int i;
    for ( i = 0; i < choiceStepsEntries; i++ ) {
	if ( numc == choiceSteps[i] ) {
	    cstep = i;
	    break;
	}
    }
    return cstep;
}
int OpenResultFile::errorToEStep( double error ) {
    int estep = -1;
    unsigned int i;
    for ( i = 0; i < errorStepsEntries; i++ ) {
	if ( error == errorSteps[i] ) {
	    estep = i;
	    break;
	}
    }
    return estep;
}

int flushOpenResultFile( OpenResultFile* orf ) {
    ResultFile* rf = orf->rf;
    return msync( rf, 0, MS_SYNC );
}

void votingSystemArrayToNameBlock( NameBlock* ret, VotingSystem** systems, int nsys ) {
    int len = 2;
    int i;
    char* block;
    for ( i = 0; i < nsys; i++ ) {
	len += strlen( systems[i]->name ) + 1;
    }
    block = (char*)malloc(len);
    ret->block = block;
    ret->blockLen = len;
    ret->names = (char**)malloc(sizeof(char*) * (nsys + 1) );
    ret->nnames = nsys;
    for ( i = 0; i < nsys; i++ ) {
	char* name;
	ret->names[i] = block;
	name = systems[i]->name;
	while ( *name ) {
	    *block = *name;
	    name++;
	    block++;
	}
	*block = *name;
	block++;
    }
    *block = '\0';
    ret->names[i] = NULL;
}

void parseNameBlock( NameBlock* it ) {
    int i;
    int nnames = 0;
    for ( i = 0; i < it->blockLen; i++ ) {
	if ( it->block[i] == '\0' ) {
	    nnames++;
	    if ( (i + 1 < it->blockLen) && (it->block[i+1] == '\0') ) {
		break;
	    }
	}
    }
    it->nnames = nnames;
    it->names = (char**)malloc( sizeof(char*) * (nnames + 1));
    it->names[0] = it->block;
    nnames = 0;
    for ( i = 0; i < it->blockLen; i++ ) {
	if ( it->block[i] == '\0' ) {
	    nnames++;
	    if ( (i + 1 < it->blockLen) && (it->block[i+1] == '\0') ) {
		break;
	    }
	    it->names[nnames] = it->block + i + 1;
	}
    }
    it->names[nnames] = NULL;
}

Result* newResult( int nsys ) {
    Result* toret = (Result*)malloc( 8 + nsys * sizeof(struct SystemResult) );
    toret->trials = 0;
    for ( int i = 0; i < nsys; i++ ) {
	toret->systems[i].meanHappiness = 0;
	toret->systems[i].reliabilityStd = 0;
	toret->systems[i].consensusStdAvg = 0;
    }
    return toret;
}

void mergeResult( Result* a, Result* b, int nsys ) {
    int j;
    u_int32_t tsum = a->trials + b->trials;
    for ( j = 0; j < nsys; j++ ) {
	a->systems[j].meanHappiness = (a->systems[j].meanHappiness * a->trials + b->systems[j].meanHappiness * b->trials) / tsum;
	a->systems[j].reliabilityStd = (a->systems[j].reliabilityStd * a->trials + b->systems[j].reliabilityStd * b->trials) / tsum;
	a->systems[j].consensusStdAvg = (a->systems[j].consensusStdAvg * a->trials + b->systems[j].consensusStdAvg * b->trials) / tsum;
    }
    a->trials = tsum;
}

void printSystemResult( void* file, const struct SystemResult* it, const char* name, int style ) {
    FILE* f = (FILE*)file;
    if ( style == 2 ) {
	// slightly smaller summary
	fprintf( f, "%s\t%.9g\t%.9g\t%.9g\n", name, it->meanHappiness, it->reliabilityStd, it->consensusStdAvg );
    } else if ( style == 3 ) {
	// slightly further smaller summary
	fprintf( f, "%p\t%.9g\t%.9g\t%.9g\n", it, it->meanHappiness, it->reliabilityStd, it->consensusStdAvg );
    } else if ( style != 0 ) {
	// standard summary
	fprintf( f, "%s\tmean happiness\t%.9g\tstd\t%.9g\tavg std\t%.9g\n", name, it->meanHappiness, it->reliabilityStd, it->consensusStdAvg );
    }
}

void printResult( void* file, const Result* r, char** names, int nsys, int style ) {
    int i;
    for ( i = 0; i < nsys; i++ ) {
	printSystemResult( file, r->systems + i, names ? names[i] : NULL, style );
    }
}

void printResultHeader( void* file, u_int32_t numc, u_int32_t numv, u_int32_t trials, float error, int nsys ) {
    FILE* f = (FILE*)file;
    if ( error < 0 ) {
        fprintf( f, "%d candidates, %d voters, %d trial elections, %d systems\n",
            numc, numv, trials, nsys );
    } else {
        fprintf( f, "%d candidates, %d voters, %d trial elections, +/- %f confusion error, %d systems\n",
            numc, numv, trials, error, nsys );
    }
}
