#ifndef RESULT_FILE_H
#define RESULT_FILE_H

#include <sys/types.h>

// struct on disk
struct SystemResult {
    double meanHappiness; // average
    double reliabilityStd; // stddev of averages
    double consensusStdAvg; // avegage of stddevs
};

// struct on disk
struct Result {
    u_int32_t trials;
    struct SystemResult systems[];
};
typedef struct Result Result;

#define sizeofRusult( nsys ) (sizeof(struct SystemResult) * nsys + sizeof(u_int32_t) * 2)

Result* newResult( int nsys );
// a = weighted average of a and b
void mergeResult( Result* a, Result* b, int nsys );
void printSystemResult( void* file, const struct SystemResult* it, const char* name, int style = 1 );
void printResult( void* file, const Result* it, char** names, int nsys, int style = 1 );
void printResultHeader( void* file, u_int32_t numc, u_int32_t numv, u_int32_t trials, float error, int nsys );

// struct on disk
struct ResultFile {
    // geometry
    u_int32_t voterStepsOffset;
    u_int32_t voterStepsEntries;
    u_int32_t choiceStepsOffset;
    u_int32_t choiceStepsEntries;
    u_int32_t errorStepsOffset;
    u_int32_t errorStepsEntries;
    u_int32_t systemsOffset; // last thing, ragged bunch of c-strings
    u_int32_t systemsEntries;
    
    // big array, loop order v, c, e, s
    u_int32_t dataStartOffset;
    u_int32_t pad; // make 8 byte aligned, like Result
};
typedef struct ResultFile ResultFile;

struct NameBlock {
    char* block;
    int blockLen;
    char** names;
    int nnames;
};
typedef struct NameBlock NameBlock;

class VotingSystem;

// struct in ram, at runtime
class OpenResultFile {
public:
    ResultFile* rf;
    size_t rfSize;
    char* path;
    int fd;

    u_int32_t* voterSteps;
    u_int32_t* choiceSteps;
    float* errorSteps;
    u_int32_t voterStepsEntries;
    u_int32_t choiceStepsEntries;
    u_int32_t errorStepsEntries;
    u_int32_t systemsEntries;

    int vstep, cstep, estep;

    NameBlock names;
    
    // pointer into mmap()ed space
    caddr_t data;
    inline Result* resultByIndex( int v, int c, int e ) {
	return (Result*)(data + ((v * vstep) + (c * cstep) + (e * estep)));
    }
    
    OpenResultFile();
    int open( char* path );
    int create( char* path, VotingSystem** s, int ns,
	u_int32_t* voterStepsIn,
	u_int32_t voterStepsEntriesIn,
	u_int32_t* choiceStepsIn,
	u_int32_t choiceStepsEntriesIn,
	float* errorStepsIn,
	u_int32_t errorStepsEntriesIn );
    int flush();
    int close();
    int votersToVStep( u_int32_t voters );
    int choicesToCStep( u_int32_t choices );
    int errorToEStep( double error );
};

#if __cplusplus
extern "C" {
#endif

// *Entries inited to how big file will be.
// on success all other fields filled in
int newOpenResultFile( OpenResultFile* orf );
int flushOpenResultFile( OpenResultFile* orf );
int closeOpenResultFile( OpenResultFile* orf );

#if __cplusplus
}
#endif

#if ! __cplusplus
__inline Result* resultByIndex( OpenResultFile* base, int v, int c, int e ) {
    return (Result*)(base->data + ((v * base->vstep) + (c * base->cstep) + (e * base->estep)));
}
#else
inline Result* resultByIndex( OpenResultFile* base, int v, int c, int e ) {
    return base->resultByIndex( v, c, e );
}
#endif

void votingSystemArrayToNameBlock( NameBlock* ret, VotingSystem** systems, int nsys );
void parseNameBlock( NameBlock* it );
void makeBlock( NameBlock* names );

#endif

