#include "WorkQueue.h"

#include <stdio.h>

WorkUnit::WorkUnit( int nv, int nc, float e, int t )
    : numv( nv ), numc( nc ), error( e ), trials( t ), next( (WorkUnit*)0 )
{}
void WorkUnit::print( void* file ) {
    FILE* f = (FILE*)file;
    fprintf( f, "%d voters, %d choices, %0.2f error, %d trials\n", numv, numc, error, trials );
}

WorkQueue::WorkQueue()
    : head( NULL ), tail( NULL )
{}
WorkQueue::~WorkQueue() {
    WorkUnit* cur = head;
    while ( cur ) {
	WorkUnit* t;
	t = cur->next;
	delete cur;
	cur = t;
    }
}

void WorkQueue::add( WorkUnit* it ) {
    it->next = NULL;
    if ( tail == NULL ) {
	head = tail = it;
    } else {
	tail->next = it;
	tail = it;
    }
}

WorkUnit* WorkQueue::pop() {
    if ( head == NULL ) {
	return NULL;
    } else {
	WorkUnit* toret = head;
	head = head->next;
	if ( head == NULL ) tail = NULL;
	return toret;
    }
}

WorkUnit* WorkQueue::newWorkUnit() {
    return pop();
}

class Steps : public WorkSource {
public:
    virtual WorkUnit* newWorkUnit();
protected:
    const int* vsteps;
    int vstepsLen;
    const int* csteps;
    int cstepsLen;
    const float* esteps;
    int estepsLen;
    int n;
    int v, c, e;
public:
    Steps();
    Steps( const char* vlist, const char* clist, const char* elist, int nIn );
};

Steps::Steps()
    : v( 0 ), c( 0 ), e( -1 )
{
    static const int dvsteps[] = { 100, 1000 };
    static const int dcsteps[] = { 2, 3, 4, 5, 6, 7, 8 };
    static const float desteps[] = { -1, 0.00, 0.01, 0.1, 0.25, 0.5, 1.0, 1.5, 2.0 };
    vsteps = dvsteps;
    vstepsLen = 2;
    csteps = dcsteps;
    cstepsLen = 7;
    esteps = desteps;
    estepsLen = 9;
    n = 10000;
}

#include <ctype.h>
#include <string.h>
#include <stdlib.h>

Steps::Steps( const char* vlist, const char* clist, const char* elist, int nIn )
    : vstepsLen( 1 ), cstepsLen( 1 ), estepsLen( 1 ),
      n( nIn ), v( 0 ), c( 0 ), e( -1 )
{
    const char* cur;
    int pos;

    cur = vlist;
    while ( *cur != '\0' ) {
	if ( isspace( *cur )) {
	    vstepsLen++;
	    do {
		cur++;
	    } while ( isspace( *cur ));
	} else {
	    cur++;
	}
    }
    vsteps = new int[vstepsLen];
    cur = vlist;
    pos = 0;
    ((int*)vsteps)[pos] = atoi( cur );
    while ( *cur != '\0' ) {
	if ( isspace( *cur )) {
	    pos++;
	    do {
		cur++;
	    } while ( isspace( *cur ));
	    ((int*)vsteps)[pos] = atoi( cur );
	}
	cur++;
    }
    cur = clist;
    while ( *cur != '\0' ) {
	if ( isspace( *cur )) {
	    cstepsLen++;
	}
	cur++;
    }
    csteps = new int[cstepsLen];
    cur = clist;
    pos = 0;
    ((int*)csteps)[pos] = atoi( cur );
    while ( *cur != '\0' ) {
	if ( isspace( *cur )) {
	    pos++;
	    do {
		cur++;
	    } while ( isspace( *cur ));
	    ((int*)csteps)[pos] = atoi( cur );
	}
	cur++;
    }
    cur = elist;
    while ( *cur != '\0' ) {
	if ( isspace( *cur )) {
	    printf("espace '%c'\n", *cur );
	    estepsLen++;
	}
	cur++;
    }
    esteps = new float[estepsLen];
    cur = elist;
    pos = 0;
    ((float*)esteps)[pos] = atof( cur );
    while ( *cur != '\0' ) {
	if ( isspace( *cur )) {
	    pos++;
	    do {
		cur++;
	    } while ( isspace( *cur ));
	    ((float*)esteps)[pos] = atof( cur );
	}
	cur++;
    }
}


WorkUnit* Steps::newWorkUnit() {
    e++;
    if ( e == estepsLen ) {
	e = 0;
	c++;
    }
    if ( c == cstepsLen ) {
	c = 0;
	v++;
    }
    if ( v == vstepsLen ) {
	v = 0;
	return NULL;
    }
    WorkUnit* toret = new WorkUnit( vsteps[v], csteps[c], esteps[e], n );
    return toret;
}

#if STEPS_TEST_MAIN
int main( int argc, char** argv ) {
    Steps* s;
    s = new Steps( argv[1], argv[2], argv[3], 1000 );
    WorkUnit* wu;
    while ( (wu = s->newWorkUnit()) != NULL ) {
	wu->print( stdout );
	delete wu;
    }
}
#endif

ThreadSafeWorkSource::ThreadSafeWorkSource( WorkSource* source )
    : realSource( source )
{
    pthread_mutex_init( &m, NULL );
}
ThreadSafeWorkSource::~ThreadSafeWorkSource() {
    pthread_mutex_destroy( &m );
}

WorkUnit* ThreadSafeWorkSource::newWorkUnit() {
    WorkUnit* toret;
    pthread_mutex_lock( &m );
    toret = realSource->newWorkUnit();
    pthread_mutex_unlock( &m );
    return toret;
}
