#ifndef RATED_VOTING_SYSTEM_H
#define RATED_VOTING_SYSTEM_H
/**
 * Interface for any voting system that inputs a voter's rankings of the choices.
 * Possible implementations: Borda count, Condorcet, IRV (sucks).
 */

#include "VotingSystem.h"

class RatedVotingSystem : public VotingSystem {
public:
    /* In addition to interfaced methods, there shall be a constructor taking 
     * the integer number N of choices to be voted on. */
    RatedVotingSystem( int numCandidates ) : VotingSystem( numCandidates ) {};
    /**
     * @param ranking An array int[N].
     *	ranking[i] is the ranking of choice i.
     */
    virtual int voteRating( int* rating ) = 0;

    /**
     * Convert to Rating and pass to voteRating();
     * @param ranking An array int[N].
     *	ranking[i] is the ranking of choice i.
     */
    virtual int voteRanking( int* ranking );
    
    /**
     * Return the number of bytes needed to represent a state save.
     */
//    virtual size_t size() = 0;
    
    /**
     * Uses write(2) to save state to the file descriptor fd.
     * @returns number of bytes written or -1 on error.
     */
    virtual int write( int fd ) = 0;

    /**
     * Uses read(2) to restore state from the file descriptor fd.
     * @returns number of bytes read or -1 on error.
     */
    virtual int read( int fd ) = 0;
};

#endif
