#ifndef VOTING_SYSTEM_H
#define VOTING_SYSTEM_H
/**
 * Interface for any voting system that inputs a voter's rankings of the choices.
 * Possible implementations: Borda count, Condorcet, IRV (sucks).
 */

class VotingSystem {
public:
    /* In addition to interfaced methods, there shall be a constructor taking 
     * the integer number N of choices to be voted on. */
    VotingSystem( int numCandidates ) : numc( numCandidates ) {};

    /**
     * @param rating An array int[N].
     *	rating[i] is the rating of choice i.
     *  A rating can be any int value. Higher means more preferred.
     *  0 for unrated.
     */
    virtual int voteRating( int* rating ) = 0;
    /**
     * @param ranking An array int[N].
     *	ranking[i] is the ranking of choice i.
     *  rankings are 1 (most preferred) through N (least).
     *  INT_MAX (or N) for unspecified.
     */
    virtual int voteRanking( int* ranking ) = 0;
    
    /**
     * @param choiceIndecies choiceIndecies[0] through choiceIndecies[number of tied winners - 1]
     *	will be set to the index [0..N-1] of the winning choice(s). Other positions undefined.
     * @return number of tied winners (hopefully 1)
     */
    virtual int getWinners( int* choiceIndecies ) = 0;
    
    /**
     * Return the number of bytes needed to represent a state save.
     */
//    virtual size_t size() = 0;
    
    /**
     * Uses write(2) to save state to the file descriptor fd.
     * @returns number of bytes written or -1 on error.
     */
    virtual int write( int fd ) = 0;

    /**
     * Uses read(2) to restore state from the file descriptor fd.
     * @returns number of bytes read or -1 on error.
     */
    virtual int read( int fd ) = 0;

protected:
    /** number of candidates */
    int numc;
};

#endif
