package org.bolson.vote;
/**
 * Interface for any voting system that inputs a voter's rankings of the choices.
 * Possible implementations: Borda count, Condorcet, IRV (sucks).
 */

public abstract class VotingSystem implements java.io.Serializable {
    /** number of candidates */
    protected int numc;

    /* In addition to interfaced methods, there shall be a constructor taking 
     * the integer number N of choices to be voted on. */
    public VotingSystem( int numCandidates ) {
	numc = numCandidates;
    };

    /**
     * @param rating An array int[N].
     *	rating[i] is the rating of choice i.
     *  A rating can be any int value. Higher means more preferred.
     *  0 for unrated.
     */
    public abstract int voteRating( int rating[] );
    /**
     * @param ranking An array int[N].
     *	ranking[i] is the ranking of choice i.
     *  rankings are 1 (most preferred) through N (least).
     *  INT_MAX (or N) for unspecified.
     */
    public abstract int voteRanking( int ranking[] );
    
    /**
     * @return indecies of winners (hopefully 1 of them)
     */
    public abstract int[] getWinners();
}
