package org.bolson.vote;
import java.io.*;

public class vote {
static void voteFromTextFile( String filename, Class fac, int numc, boolean rating, boolean isFloat, boolean isBin )  throws Exception {
    VotingSystem tab;
    int p[];
    StreamTokenizer st = null;
    DataInputStream dis = null;
    boolean goodP = true;
    int votes = 0;

//    System.out.println("reading votes from \"" + filename + "\", processing with " + fac + ", " + numc + " choices, " + (rating ? "Rating " : "Ranking " ) + (isFloat ? "float " : "int ") + (isBin ? "bin" : "text") );
    if ( isBin ) {
	dis = new DataInputStream( new FileInputStream( filename ));
	if ( numc == -1 ) numc = dis.readInt();
    } else {
	st = new StreamTokenizer( new FileReader( filename ));
	if ( numc == -1 ) { if ( st.nextToken() != st.TT_NUMBER ) return; numc = (int)st.nval; }
    }
    tab = (VotingSystem)(fac.getConstructors())[0].newInstance( new Object[]{new Integer( numc )} );
    p = new int[numc];
    try { while ( goodP ) {
//	System.out.print("vote:");
	for ( int i = 0; i < numc; i++ ) {
	    int err;
	    if ( isBin ) {
		if ( isFloat ) {
		    float tf;
		    tf = dis.readFloat();
		    p[i] = (int)(tf * Integer.MAX_VALUE);
		} else {
		    p[i] = dis.readInt();
		}
	    } else {
		if ( st.nextToken() != st.TT_NUMBER ) {
		    goodP = false;
		    break;
		}
		if ( isFloat ) {
		    p[i] = (int)(st.nval * Integer.MAX_VALUE);
		} else {
		    p[i] = (int)(st.nval);
		}
	    }
//	    System.out.print( "\t"+ p[i] );
	}
//	System.out.println();
	if ( goodP ) {
	    if ( rating ) {
		tab.voteRating( p );
	    } else {
		tab.voteRanking( p );
	    }
	    votes++;
	}
    }} catch ( Exception e ) {
	e.printStackTrace();
    };
    System.out.println( votes + " votes counted" );
    int[] winners = tab.getWinners();
    if ( winners.length == 0 ) {
	System.out.println("no winner found,"+tab);
    } else if ( winners.length == 1 ) {
	System.out.println("winner, choice number " + (winners[0] + 1) + " (of 1.." + numc + ")" );
    } else {
	System.out.print( winners.length + " winners, numbers " + (winners[0] + 1) );
	for ( int i = 1; i < winners.length; i++ ) {
	    System.out.print(", " + (winners[i] + 1) );
	}
	System.out.println(" (of 1.." + numc + ")\n" );
    }
}

static final String optstring = "bBcdDfhi:n:Rrw";
static final String usageText = "vote [-bBcdfhRrw][-n number of choices][-i vote file][vote file ...]\n";
static final String helpText =
"\t-b\tbinary\n\t-B\ttext\n\t-c\tCondorcet\n\t-D\tBorda count\n\t-d\tinteger\n\t-f\tfloat (-1.0 .. 1.0)\n\t-h\tthis help\n\t-i f\timmediately load and run vote file f\n\t-n N\tnumber of choices\n\t-R\tnumbers are rankings (1 best)\n\t-r\tnumbers are ratings (higher better)\n\t-w\tRaw Sum of ratings\n";

public static void main( String argv[] ) {
    int c;
    int i;
    
    boolean rating = true;
    boolean isFloat = true;
    boolean isBin = false;
    Class fac = null;
    int numc = -1;

    try {
	fac = Class.forName("org.bolson.vote.RawRating");
    } catch ( java.lang.ClassNotFoundException e ) {
	e.printStackTrace();
    }
    for ( i = 0; i < argv.length; i++ ) {
	if ( argv[i].equals("-B") ) {
	    isBin = false;
	} else if ( argv[i].equals("-b") ) {
	    isBin = true;
	} else if ( argv[i].equals("-c") ) {
	    try {
		fac = Class.forName("org.bolson.vote.Condorcet");
	    } catch ( java.lang.ClassNotFoundException e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("-D") ) {
	    try {
		fac = Class.forName("org.bolson.vote.BordaVotingSystem");
	    } catch ( java.lang.ClassNotFoundException e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("-d") ) {
	    isFloat = false;
	} else if ( argv[i].equals("-f") ) {
	    isFloat = true;
	} else if ( argv[i].equals("-h") ) {	// help
	    System.out.println(usageText);
	    System.out.println(helpText);
	} else if ( argv[i].equals("-i") ) {
	    i++;
	    try {
		voteFromTextFile( argv[i], fac, numc, rating, isFloat, isBin );
	    } catch ( Exception e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("-n") ) {
	    i++;
	    numc = Integer.parseInt( argv[i] );
	} else if ( argv[i].equals("-R") ) {
	rating = false;
	} else if ( argv[i].equals("-r") ) {
	rating = true;
	} else if ( argv[i].equals("-w") ) {
	    try {
		fac = Class.forName("org.bolson.vote.RawRating");
	    } catch ( java.lang.ClassNotFoundException e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("--system") ) {
	    i++;
	    try {
		fac = Class.forName(argv[i]);
	    } catch ( java.lang.ClassNotFoundException e ) {
		e.printStackTrace();
	    }
	} else if ( argv[i].equals("--") ) {
	    break;
	} else if ( argv[i].startsWith("-") ) {
	    System.err.println("bogus flag: " + argv[i] );
	    return;
	} else {
	    break;
	}
    }

    for ( ; i < argv.length; i++ ) {
	try {
	    voteFromTextFile( argv[i], fac, numc, rating, isFloat, isBin );
	} catch ( Exception e ) {
	    e.printStackTrace();
	}
    }
}
}
