/* This file is placed in the public domain without warrantee by it's creator, Brian Olson. */
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <time.h>
#include <unistd.h>
#include <math.h>
#include <string.h>

#include "Condorcet.h"
#include "BordaVotingSystem.h"
#include "RawRating.h"

class runParams {
public:
    bool rating;
    bool isFloat;
    bool isBin;
    bool scriptMode;
    VotingSystem* (*fac)(int);
    FILE* fin;
    int numc;
    int numWinners;
    int* p;
    
    runParams();
};

runParams::runParams()
    : rating( false ), isFloat( false ), isBin( false ), scriptMode( false ),
      fac( BordaVotingSystem::newBordaVotingSystem ), fin( NULL ), numc( -1 ),
      numWinners( -1 ), p( NULL )
{}

int voteFromTextFile( runParams* r ) {
    VotingSystem* tab;
    int* p;
    int numc = r->numc;
    if ( !r->fin ) return -1;
    if ( r->isBin ) {
	if ( numc == -1 ) fread( &numc, sizeof(numc), 1, r->fin );
    } else {
	if ( numc == -1 ) fscanf( r->fin, "%d", &numc );
    }
    tab = r->fac( numc );
    p = new int[numc];
    while ( 1 ) {
//	printf("voting %s\t", rating ? "rating " : "ranking" );
	for ( int i = 0; i < numc; i++ ) {
	    int err;
	    if ( r->isBin ) {
		if ( r->isFloat ) {
		    float tf;
		    err = fread( &tf, sizeof(float), 1, r->fin );
		    p[i] = (int)(tf * INT_MAX);
		} else {
		    err = fread( &(p[i]), sizeof(int), 1, r->fin );
		}
	    } else {
		if ( r->isFloat ) {
		    float tf;
		    err = fscanf( r->fin, "%f", &tf );
		    p[i] = (int)(tf * INT_MAX);
		} else {
		    err = fscanf( r->fin, "%d",&(p[i]));
		}
	    }
	    if ( err <= 0 ) goto noMoreVotes;
//	    printf("%d\t", p[i] );
	}
//	printf("\n");
	if ( r->rating ) {
	    tab->voteRating( p );
	} else {
	    tab->voteRanking( p );
	}
    }
  noMoreVotes:
    fclose( r->fin );
    r->fin = NULL;
    r->numWinners = tab->getWinners( p );
    r->p = p;
    delete tab;
    return 0;
}

void printWinners( runParams* r ) {
    if ( r->numWinners == 1 ) {
	printf("winner, choice number %d (of 1..%d)\n", r->p[0] + 1, r->numc );
    } else {
	printf("%d winners, numbers %d", r->numWinners, r->p[0] + 1 );
	for ( int i = 1; i < r->numWinners; i++ ) {
	    printf(", %d", r->p[i] + 1 );
	}
	printf(" (of 1..%d)\n", r->numc );
    }
    delete [] r->p;
    r->p = NULL;
}

char* optstring = "bBcdDfhi:n:Rrsw";
const char* const usageText = "vote [-bBcdfhRrw][-n number of choices][-i vote file][vote file ...]\n";
const char* const helpText =
"\t-b\tbinary\n"
"\t-B\ttext (default)\n"
"\t-c\tCondorcet\n"
"\t-D\tBorda count (default)\n"
"\t-d\tinteger (default)\n"
"\t-f\tfloat (-1.0 .. 1.0)\n"
"\t-h\tthis help\n"
"\t-i f\timmediately load and run vote file f\n"
"\t-n N\tnumber of choices\n"
"\t-R\tnumbers are rankings (1 best) (default)\n"
"\t-r\tnumbers are ratings (higher better)\n"
"\t-s\tscript friendly output\n"
"\t-w\tRaw Sum of ratings\n"
;

FILE* myOpen( const char* optarg ) {
    if ( !strcmp(optarg,"-") ) {
	return stdin;
    } else {
	FILE* fin = fopen( optarg, "r" );
	if ( !fin ) {
	    perror( optarg );
	}
	return fin;
    }
}


int main( int argc, char** argv ) {
    int c;
    extern char* optarg;
    extern int optind;
    
    runParams r;
    
    while ( (c = getopt(argc,argv,optstring)) != -1 ) switch ( c ) {
    case 'B':
	r.isBin = false;
	break;
    case 'b':
	r.isBin = true;
	break;
    case 'c':
	r.fac = Condorcet::newCondorcet;
	break;
    case 'D':
	r.fac = BordaVotingSystem::newBordaVotingSystem;
	break;
    case 'd':
	r.isFloat = false;
	break;
    case 'f':
	r.isFloat = true;
	break;
    case 'h':	// help
	printf(usageText);
	printf(helpText);
	break;
    case 'i':
	r.fin = myOpen( optarg );
    	r.numWinners = voteFromTextFile( &r );
	printWinners( &r );
	break;
    case 'n':
	r.numc = atoi( optarg );
	break;
    case 'R':
	r.rating = false;
	break;
    case 'r':
	r.rating = true;
	break;
    case 's':
	r.scriptMode = true;
	break;
    case 'w':
	r.fac = RawRating::newRawRating;
	break;
    default:
	printf("unknown option \'%c\'\n", c );
	return 1;
    }

    for ( ; optind < argc; optind++ ) {
	optarg = argv[optind];
	r.fin = myOpen( optarg );
    	voteFromTextFile( &r );
	printWinners( &r );
    }

    return 0;
}
