#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <time.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#if 0
#include "Voter.h"
#include "VoterSim.h"
#include "VotingSystem.h"
#endif

#include "ResultFile.h"
#include "DBResultFile.h"

const char* outputModes[][2] = {
   {"set terminal png color\nset size 1.5,1.5\n",".png"},
   {"set terminal postscript landscape color solid\n",".ps"},
   {"set terminal postscript eps color solid\n",".eps"},
   { NULL, NULL }
};



#define MAX_STEPS 100
int csteps[MAX_STEPS];
int vsteps[MAX_STEPS];
float esteps[MAX_STEPS];
int numCsteps = 0, numVsteps = 0, numEsteps = 0;

// a DBT is small (2 machine words), have a lot of them.
#define MAX_KEYS 8000
//DBT keys[MAX_KEYS];
int numKeys = 0;

inline int present( int v, int* arr, int len ) {
    for ( int i = 0; i < len; i++ ) {
	if ( arr[i] == v ) {
	    return 1;
	}
    }
    return 0;
}
inline int present( float v, float* arr, int len ) {
    for ( int i = 0; i < len; i++ ) {
	if ( arr[i] == v ) {
	    return 1;
	}
    }
    return 0;
}
inline void insert( int v, int* arr, int& len ) {
    int i = len - 1;
    len++;
    for ( ; i >= 0; i-- ) {
	if ( arr[i] > v ) {
	    arr[i+1] = arr[i];
	} else {
	    arr[i+1] = v;
	    return;
	}
    }
    arr[0] = v;
}
inline void insert( float v, float* arr, int& len ) {
    int i = len - 1;
    len++;
    for ( ; i >= 0; i-- ) {
	if ( arr[i] > v ) {
	    arr[i+1] = arr[i];
	} else {
	    arr[i+1] = v;
	    return;
	}
    }
    arr[0] = v;
}

char* outputPrefix = "graph/";
DBResultFile* drf;

void byvc( FILE* GP, int v, int c, const char* omode[2] ) {
    char outputname[256];
    Result* r;
    int err;

    fprintf( GP, omode[0] );
    snprintf( outputname, sizeof(outputname), "%sv%d", outputPrefix, v );
    err = mkdir( outputname, 0755 );
    if ( err == -1 && errno != EEXIST ) {
	perror("mkdir");
	return;
    }
    snprintf( outputname, sizeof(outputname), "%sv%d/c%02d%s", outputPrefix, v, c, omode[1] );
    fprintf( GP, "set output \"%s\"\n", outputname );
    fprintf( GP, "set nologscale x\n" );
    fprintf( GP, "set data style linespoints\n" );
    fprintf( GP, "set xlabel \"Potential Error\"\n" );
    fprintf( GP, "set ylabel \"Happiness\"\n" );
    fprintf( GP, "set title \"%d Choices, %d Voters\"\n", c, v );
    fprintf( GP, "plot \\\n" );
    for ( int hi = 0; hi < drf->names.nnames; hi++ ) {
	if ( hi + 1 >= drf->names.nnames ) {
	    fprintf( GP, "\'-\' title \"%s\"\n", drf->names.names[hi] );
	} else {
	    fprintf( GP, "\'-\' title \"%s\", \\\n", drf->names.names[hi] );
	}
    }
    for ( int hi = 0; hi < drf->names.nnames; hi++ ) {
	// print all data with v and c, and various e
	for ( int ei = 0; ei < numEsteps; ei++ ) {
	    r = drf->get( c, v, esteps[ei] );
	    if ( r ) {
		fprintf( GP, "%f\t%f\n", (esteps[ei] == -1.00) ? -.1 : esteps[ei], r->systems[hi].meanHappiness );
	    }
	}
	fprintf( GP, "e\n");
    }
    fprintf( GP, "\n" );
}

void byve( FILE* GP, int v, float e, const char* omode[2] ) {
    char outputname[256];
    Result* r;
    int err;

    fprintf( GP, omode[0] );
    snprintf( outputname, sizeof(outputname), "%sv%d", outputPrefix, v );
    err = mkdir( outputname, 0755 );
    if ( err == -1 && errno != EEXIST ) {
	perror("mkdir");
	return;
    }
    snprintf( outputname, sizeof(outputname), "%sv%d/e%0.2f%s", outputPrefix, v, e, omode[1] );
    fprintf( GP, "set output \"%s\"\n", outputname );
    fprintf( GP, "set logscale x\n" );
    fprintf( GP, "set data style linespoints\n" );
    fprintf( GP, "set xlabel \"Number of Choices\"\n" );
    fprintf( GP, "set ylabel \"Happiness\"\n" );
    fprintf( GP, "set title \"Potential Error = %0.2f, %d Voters\"\n", e, v );
    fprintf( GP, "plot \\\n" );
    for ( int hi = 0; hi < drf->names.nnames; hi++ ) {
	if ( hi + 1 >= drf->names.nnames ) {
	    fprintf( GP, "\'-\' title \"%s\"\n", drf->names.names[hi] );
	} else {
	    fprintf( GP, "\'-\' title \"%s\", \\\n", drf->names.names[hi] );
	}
    }
    for ( int hi = 0; hi < drf->names.nnames; hi++ ) {
	// print all data with v and c, and various e
	for ( int ci = 0; ci < numCsteps; ci++ ) {
	    r = drf->get( csteps[ci], v, e );
	    if ( r ) {
		fprintf( GP, "%d\t%f\n", csteps[ci], r->systems[hi].meanHappiness );
	    }
	}
	fprintf( GP, "e\n");
    }
    fprintf( GP, "\n" );
}
void byce( FILE* GP, int c, float e, const char* omode[2] ) {
    char outputname[256];
    Result* r;

    fprintf( GP, omode[0] );
    snprintf( outputname, sizeof(outputname), "%sc%02de%0.2f%s", outputPrefix, c, e, omode[1] );
    fprintf( GP, "set output \"%s\"\n", outputname );
    fprintf( GP, "set logscale x\n" );
    fprintf( GP, "set data style linespoints\n" );
    fprintf( GP, "set xlabel \"Number of Voters\"\n" );
    fprintf( GP, "set ylabel \"Happiness\"\n" );
    fprintf( GP, "set title \"Potential Error = %0.2f, %d Choices\"\n", e, c );
    fprintf( GP, "plot \\\n" );
    for ( int hi = 0; hi < drf->names.nnames; hi++ ) {
	if ( hi + 1 >= drf->names.nnames ) {
	    fprintf( GP, "\'-\' title \"%s\"\n", drf->names.names[hi] );
	} else {
	    fprintf( GP, "\'-\' title \"%s\", \\\n", drf->names.names[hi] );
	}
    }
    for ( int hi = 0; hi < drf->names.nnames; hi++ ) {
	// print all data with v and c, and various e
	for ( int vi = 0; vi < numVsteps; vi++ ) {
	    r = drf->get( c, vsteps[vi], e );
	    if ( r ) {
		fprintf( GP, "%d\t%f\n", vsteps[vi], r->systems[hi].meanHappiness );
	    }
	}
	fprintf( GP, "e\n");
    }
    fprintf( GP, "\n" );
}

int main( int argc, char** argv, char** envp ) {
    drf = DBResultFile::open( argv[1] );
    {
	DBT k, d;
	int err;
	k.data = NULL;
	k.size = 0;
	err = drf->db->seq( drf->db, &k, &d, R_FIRST );
	while ( err == 0 ) {
	    //printf("%0.8x\t%0.8x\n", ((u_int32_t*)k.data)[0], ((u_int32_t*)k.data)[1] );
	    //hexprint( k.data, k.size );
	    if ( ((u_int32_t*)k.data)[0] <= 0x00ffffff ) {
//		Result* r = (Result*)d.data;
		ResultKeyStruct* rks = (ResultKeyStruct*)k.data;
//		printResultHeader( stdout, rks->choices, rks->voters, r->trials, rks->error, drf->numSystems() );
//		printResult( stdout, r, drf->names.names, drf->numSystems() );
		if ( ! present( rks->choices, csteps, numCsteps ) ) {
		    insert( rks->choices, csteps, numCsteps );
		}
		if ( ! present( rks->voters, vsteps, numVsteps ) ) {
		    insert( rks->voters, vsteps, numVsteps );
		}
		if ( ! present( rks->error, esteps, numEsteps ) ) {
		    insert( rks->error, esteps, numEsteps );
		}
	//	keys[numKeys] = k;
		numKeys++;
	    }
	    k.data = NULL;
	    k.size = 0;
	    err = drf->db->seq( drf->db, &k, &d, R_NEXT );
	}
    }

    printf("%d keys\n", numKeys);
    int i;
    printf("vsteps:");
    for ( i = 0; i < numVsteps; i++ ) {
	printf(" %d", vsteps[i] );
    }
    printf("\n");
    printf("csteps:");
    for ( i = 0; i < numCsteps; i++ ) {
	printf(" %d", csteps[i] );
    }
    printf("\n");
    printf("esteps:");
    for ( i = 0; i < numEsteps; i++ ) {
	printf(" %f", esteps[i] );
    }
    printf("\n");
    FILE* GP = fopen( "gnuplotCommands", "w" );
    for ( int mi = 0; outputModes[mi][0] != NULL; mi++ ) {
	for ( int vi = 0; vi < numVsteps; vi++ ) {
	    for ( int ci = 0; ci < numCsteps; ci++ ) {
		byvc( GP, vsteps[vi], csteps[ci], outputModes[mi] );
	    }
	    for ( int ei = 0; ei < numEsteps; ei++ ) {
		byve( GP, vsteps[vi], esteps[ei], outputModes[mi] );
	    }
	}
	for ( int ci = 0; ci < numCsteps; ci++ ) {
	    for ( int ei = 0; ei < numEsteps; ei++ ) {
		byce( GP, csteps[ci], esteps[ei], outputModes[mi] );
	    }
	}
    }
//    byvc( GP, vsteps[0], csteps[0], outputModes[0] );
    fclose( GP );
    delete drf;
    system("ps -u");
    sleep(1);
}
